! ***********************************************************************
!
!   Copyright (C) 2010  Bill Paxton
!
!   this file is part of mesa.
!
!   mesa is free software; you can redistribute it and/or modify
!   it under the terms of the gnu general library public license as published
!   by the free software foundation; either version 2 of the license, or
!   (at your option) any later version.
!
!   mesa is distributed in the hope that it will be useful, 
!   but without any warranty; without even the implied warranty of
!   merchantability or fitness for a particular purpose.  see the
!   gnu library general public license for more details.
!
!   you should have received a copy of the gnu library general public license
!   along with this software; if not, write to the free software
!   foundation, inc., 59 temple place, suite 330, boston, ma 02111-1307 usa
!
! ***********************************************************************
 
      module run_star_extras

      use star_lib
      use star_def
      use const_def
      use run_star_support
      
      implicit none

      character (len=256) :: masses_filename
      
      double precision, pointer :: masses(:)
      integer :: nmasses, which_mass

      ! these routines are called by the standard run_star check_model
      contains

      ! load an array of masses from mass.list and execute star for each
      ! starting mass in the series.
      subroutine do_run
         use run_star_support
         
         type (star_info), pointer :: s
         integer :: id, ierr, i, j, result
         logical :: first_try, restart

         logical, parameter :: dbg = .false., do_init_and_alloc = .false., &
            do_free_star = .false., okay_to_restart = .true.
         
         include 'formats'
         
         ierr = 0
         
         masses_filename = 'mass.list' ! file containing list of masses for grid

         nullify(masses)
         call read_masses(masses_filename, masses, nmasses, ierr)
         if (ierr /= 0) return

         call init_and_alloc(id, s, ierr)
         if (failed('preload')) return
         
         do i=1, nmasses
         
            which_mass = i ! for use by multistar_controls
            
            call run1_star( &
               do_init_and_alloc, do_free_star, okay_to_restart, &
               id, restart, &
               multistar_controls, &
               extras_startup, &
               extras_check_model, &
               how_many_extra_history_columns, &
               data_for_extra_history_columns, &
               how_many_extra_profile_columns, &
               data_for_extra_profile_columns, &
               extras_finish_step, &
               extras_after_evolve, &
               ierr)
            if (ierr /= 0) return
            
            do j = 1, 10
               write(*,*)
            end do
                        
         end do
         
         write(*,*)

         write (*, *)
         write (*, *)
         write(*, *)
         write(*, '(a)') 'finished'
         write(*, *)
         

         call star_shutdown
                  
         contains
         
         logical function failed(str)
            character (len=*), intent(in) :: str
            failed = (ierr /= 0)
            if (failed) then
               write(*, *) trim(str) // ' ierr', ierr
               stop 1
            end if
         end function failed

      end subroutine do_run


      subroutine multistar_controls(s, ierr)
         use star_def, only: star_info
         use run_star_support, only: save_model_filename, &
            saved_model_name, create_pre_main_sequence_model, &
            load_saved_model, relax_mass, new_mass
            ! lets us set filenames for loading and saving models, and see
            ! whether we're starting from pre-ms or loading a model
         type (star_info), pointer :: s
         integer, intent(out) :: ierr
         
         integer :: i, m
         character (len=256) :: history_name, profiles_index_name, profile_data_prefix
         character (len=4) :: prefix

         include 'formats'

         ierr = 0
         i = which_mass
         s% initial_mass = masses(i)

         new_mass = masses(i) ! for relax_mass
        
         ! labeling output files with prefix according to mass, e.g.
         ! m050_history.data and m050_prems.mod for M = 0.5 Msun
         m = int(masses(i)*100)
         if (m < 10) then
            write(prefix,'(a3,i1)') 'm00', m
         else if (m < 100) then
            write(prefix,'(a2,i2)') 'm0', m
         else
            write(prefix,'(a,i3)') 'm', m
         end if
        
         write(history_name, '(a,a)') prefix, '_history.data'
         write(profiles_index_name, '(a,a)') prefix, '_profiles.data'
         write(profile_data_prefix, '(a,a)') prefix, '_profile'
         
         ! set filenames for load/save of models based on parameters
         ! in star_job namelist
         if (load_saved_model) &
            write(saved_model_name, '(a,a)') prefix, '_prems.mod'
        ! if (relax_mass) &
        !     write(saved_model_Name, '(a)') 'm010_prems.mod'
         if (create_pre_main_sequence_model .or. relax_mass) &
            write(save_model_filename, '(a,a)') prefix, '_prems.mod'

         s% star_history_name = trim(history_name)
         s% profiles_index_name = trim(profiles_index_name)
         s% profile_data_prefix = trim(profile_data_prefix)
         
         write(*,*)
         write(*,2) 'mass', i, s% initial_mass
         write(*,2) 'star_history_name: ' // trim(s% star_history_name)
         write(*,2) 'profiles_index_name: ' // trim(s% profiles_index_name)
         write(*,2) 'profile_data_prefix: ' // trim(s% profile_data_prefix)
         write(*,2) 'save_model_filename: ' // trim(save_model_filename)
         write(*,2) 'saved_model_name: ' // trim(saved_model_name)
         write(*,*)
         
      end subroutine multistar_controls      

      subroutine simple_am_mixing(id, ierr)
      ! enforces solid body rotation throughout convective regions along with
      ! angular momentum conservation for radiative regions.  probably doesn't
      ! change anything
         integer, intent(in) :: id
         integer, intent(out) :: ierr
         integer :: k, n_rad, n_conv
         type (star_info), pointer :: s
         ierr = 0

        call star_ptr(id, s, ierr)
         if (ierr /= 0) return

         n_rad = 0
         do k = 1, s% nz
            if (s% mixing_type(k) == 0) then
               s% am_nu(k) = 0
               n_rad = n_rad + 1
            end if
            if (s% mixing_type(k) == 1) s% am_nu(k) = 1e10
         end do

         n_conv = (s% nz) - n_rad

         ! write(*,*) 'radiative/convective zones: ', n_rad, n_conv

      end subroutine simple_am_mixing

      subroutine my_torque(id, ierr)
      ! null routine as is.  ultimately this is where we implement a
      ! magnetic braking law
         integer, intent(in) :: id
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         integer :: k
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         do k = 1, s% nz
            s% extra_jdot(k) = 0 ! rate at which specific angular momentum is changed
! note that can set extra_omegadot instead of extra_jdot if that is more convenient
! set one or the other, not both.  set the one you are not using to 0 as in the following line.
            s% extra_omegadot(k) = 0 ! rate at which omega is changed
         end do

      end subroutine my_torque


      
      subroutine extras_controls(s, ierr)
         type (star_info), pointer :: s
         integer, intent(out) :: ierr
         ierr = 0

         s% other_am_mixing => simple_am_mixing
         s% other_torque => my_torque
         
         ! this is the place to set any procedure pointers you want to change
         ! e.g., other_wind, other_mixing, other_energy  (see star_data.inc)
      
      
      end subroutine extras_controls      
      
      integer function extras_startup(s, id, restart, ierr)
         type (star_info), pointer :: s
         integer, intent(in) :: id
         logical, intent(in) :: restart
         integer, intent(out) :: ierr
         ierr = 0
         extras_startup = 0
         if (.not. restart) then
            call alloc_extra_info(s)
         else ! it is a restart
            call unpack_extra_info(s)
         end if
      end function extras_startup
      

      ! returns either keep_going, retry, backup, or terminate.
      integer function extras_check_model(s, id, id_extra)
         type (star_info), pointer :: s
         integer, intent(in) :: id, id_extra
         extras_check_model = keep_going         
         if (.false. .and. s% star_mass_h1 < 0.35d0) then
            ! stop when star hydrogen mass drops to specified level
            extras_check_model = terminate
            write(*, *) 'have reached desired hydrogen mass'
            return
         end if
      end function extras_check_model

   
      integer function how_many_extra_history_columns(s, id, id_extra)
         type (star_info), pointer :: s
         integer, intent(in) :: id, id_extra
         how_many_extra_history_columns = 7
      end function how_many_extra_history_columns
      
      
      subroutine data_for_extra_history_columns(s, id, id_extra, n, names, vals, ierr)
      ! adding columns to give internal rotation at a few different mass
      ! coordinates q=m/Mstar
         type (star_info), pointer :: s
         integer, intent(in) :: id, id_extra, n
         character (len=maxlen_history_column_name) :: names(n)
         real(dp) :: qs(n), vals(n)
         integer, intent(out) :: ierr

         integer :: kbdy, j
         real(dp) :: &
            m, r, lgT, lgRho, L, v, &
            lgP, g, X, Y, edv_H, edv_He, &
            scale_height, dlnX_dr, dlnY_dr, dlnRho_dr, &
            omega, omega_div_omega_crit, &
	mx1_bot_logT, mx2_bot_logT

         !note: do NOT add these names to history_columns.list
         ! the history_columns.list is only for the built-in log column options.
         ! it must not include the new column names you are adding here.
         
         ierr = 0

         call star_ptr(id, s, ierr)
         if (ierr /= 0) return

         names(1) = 'omega_q01'
         names(2) = 'omega_q05'
         names(3) = 'omega_q10'
         names(4) = 'omega_q20'
         names(5) = 'omega_q30'
         names(6) = 'omega_q50'
         names(7) = 'omega_q99'

         qs = (/ 0.01, 0.05, 0.1, 0.2, 0.3, 0.5, 0.99 /)
         
         do j = 1, n
            call star_info_at_q(s, qs(j), kbdy, m, r, lgT, lgRho, L, v, &
            lgP, g, X, Y, edv_H, edv_He, scale_height, dlnX_dr, &
            dlnY_dr, dlnRho_dr, omega, omega_div_omega_crit)
            vals(j) = omega
         end do

         if (mod(s% model_number,10) == 0) write(*,*) &
            'surface omega ', s% omega(1)

      end subroutine data_for_extra_history_columns

      
      integer function how_many_extra_profile_columns(s, id, id_extra)
         type (star_info), pointer :: s
         integer, intent(in) :: id, id_extra
         how_many_extra_profile_columns = 0
      end function how_many_extra_profile_columns
      
      
      subroutine data_for_extra_profile_columns(s, id, id_extra, n, nz, names, vals, ierr)
         type (star_info), pointer :: s
         integer, intent(in) :: id, id_extra, n, nz
         character (len=maxlen_profile_column_name) :: names(n)
         real(dp) :: vals(nz,n)
         integer, intent(out) :: ierr
         integer :: k
         ierr = 0
         
         !note: do NOT add these names to profile_columns.list
         ! the profile_columns.list is only for the built-in profile column options.
         ! it must not include the new column names you are adding here.

         ! here is an example for adding a profile column
         !if (n /= 1) stop 'data_for_extra_profile_columns'
         !names(1) = 'beta'
         !do k = 1, nz
         !   vals(k,1) = s% Pgas(k)/s% P(k)
         !end do
         
      end subroutine data_for_extra_profile_columns
      

      ! returns either keep_going or terminate.
      ! note: cannot request retry or backup; extras_check_model can do that.
      integer function extras_finish_step(s, id, id_extra)
         type (star_info), pointer :: s
         integer, intent(in) :: id, id_extra
         integer :: ierr
         extras_finish_step = keep_going
         call store_extra_info(s)

         ! to save a profile, 
            ! s% need_to_save_profiles_now = .true.
         ! to update the star log,
            ! s% need_to_update_history_now = .true.
            
      end function extras_finish_step
      
      
      subroutine extras_after_evolve(s, id, id_extra, ierr)
         type (star_info), pointer :: s
         integer, intent(in) :: id, id_extra
         integer, intent(out) :: ierr
         ierr = 0
      end subroutine extras_after_evolve
      
      
      ! routines for saving and restoring extra data so can do restarts
         
         ! put these defs at the top and delete from the following routines
         !integer, parameter :: extra_info_alloc = 1
         !integer, parameter :: extra_info_get = 2
         !integer, parameter :: extra_info_put = 3
      
      
      subroutine alloc_extra_info(s)
         integer, parameter :: extra_info_alloc = 1
         type (star_info), pointer :: s
         call move_extra_info(s,extra_info_alloc)
      end subroutine alloc_extra_info
      
      
      subroutine unpack_extra_info(s)
         integer, parameter :: extra_info_get = 2
         type (star_info), pointer :: s
         call move_extra_info(s,extra_info_get)
      end subroutine unpack_extra_info
      
      
      subroutine store_extra_info(s)
         integer, parameter :: extra_info_put = 3
         type (star_info), pointer :: s
         call move_extra_info(s,extra_info_put)
      end subroutine store_extra_info
      
      
      subroutine move_extra_info(s,op)
         integer, parameter :: extra_info_alloc = 1
         integer, parameter :: extra_info_get = 2
         integer, parameter :: extra_info_put = 3
         type (star_info), pointer :: s
         integer, intent(in) :: op
         
         integer :: i, j, num_ints, num_dbls, ierr
         
         i = 0
         ! call move_int or move_flg    
         num_ints = i
         
         i = 0
         ! call move_dbl       
         
         num_dbls = i
         
         if (op /= extra_info_alloc) return
         if (num_ints == 0 .and. num_dbls == 0) return
         
         ierr = 0
         call star_alloc_extras(s% id, num_ints, num_dbls, ierr)
         if (ierr /= 0) then
            write(*,*) 'failed in star_alloc_extras'
            write(*,*) 'alloc_extras num_ints', num_ints
            write(*,*) 'alloc_extras num_dbls', num_dbls
            stop 1
         end if
         
         contains
         
         subroutine move_dbl(dbl)
            real(dp) :: dbl
            i = i+1
            select case (op)
            case (extra_info_get)
               dbl = s% extra_work(i)
            case (extra_info_put)
               s% extra_work(i) = dbl
            end select
         end subroutine move_dbl
         
         subroutine move_int(int)
            integer :: int
            i = i+1
            select case (op)
            case (extra_info_get)
               int = s% extra_iwork(i)
            case (extra_info_put)
               s% extra_iwork(i) = int
            end select
         end subroutine move_int
         
         subroutine move_flg(flg)
            logical :: flg
            i = i+1
            select case (op)
            case (extra_info_get)
               flg = (s% extra_iwork(i) /= 0)
            case (extra_info_put)
               if (flg) then
                  s% extra_iwork(i) = 1
               else
                  s% extra_iwork(i) = 0
               end if
            end select
         end subroutine move_flg
      
      end subroutine move_extra_info

      end module run_star_extras
      
