! ***********************************************************************
!
!   Copyright (C) 2010  Bill Paxton
!
!   this file is part of mesa.
!
!   mesa is free software; you can redistribute it and/or modify
!   it under the terms of the gnu general library public license as published
!   by the free software foundation; either version 2 of the license, or
!   (at your option) any later version.
!
!   mesa is distributed in the hope that it will be useful, 
!   but without any warranty; without even the implied warranty of
!   merchantability or fitness for a particular purpose.  see the
!   gnu library general public license for more details.
!
!   you should have received a copy of the gnu library general public license
!   along with this software; if not, write to the free software
!   foundation, inc., 59 temple place, suite 330, boston, ma 02111-1307 usa
!
! ***********************************************************************
 
      module run_star_extras

      use star_lib
      use star_def
      use const_def
      
      implicit none
      
      ! these routines are called by the standard run_star check_model
      contains
      
      subroutine kawaler_jdot(id, ierr)
         	implicit double precision (a-h,o-z)
         integer, intent(in) :: id
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         integer :: k, iter
         real(dp) :: fmdot0, fk, fom, wn, fr, flum, fmtot, djdt
         
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         
         fmdot0 = 1.d0
         fk = 0.256
         fom = s% omega_avg_surf ! rad/s
         wn = 1.5
         fr = s% photosphere_r ! Rsun
         flum = s% photosphere_L ! Lsun
         fmtot = s% star_mass ! Msun

		! c calculate rate of loss of angular momentum using prescription
		! c in Kawaler (1988, ApJ 333, 236)
		! c inputs:
		! c   fmdot0 = scaling constant (in units of 10^-14 Msun/year) for wind mass loss
		! c   fk = normalization constant (use 0.256 for solar calibration)
		! c   fom = angular rotation velocity (rad/sec)
		! c   wn = magnetic field exponent (use 1.5)
		! c   fr = radius of model in solar units
		! c   flum = luminosity of model in solar units
		! c   fmtot = mass of model in solar units
		! c output:
		! c     djdt = angular momentum loss rate (cgs units)
		! c
		! c if fmdot0 is less than 0, use a constant mass loss rate
		! c of 1x10-14 solar masses per year
			  if (fmdot0.lt.0.d0) then
		! c        fmdot = -fmdot0
				fmdot=1.d0
			  else
		! c mass loss rate: modified reimers
				fmdot = fmdot0*flum*fr/fmtot
			  endif
		! c
		! c compute ra/R
		! c
			  xr=10.d0
			  xr0=0.d0
			  iter=0
			  do while (dabs(xr-xr0)/xr.gt.1.d-4.and.iter.lt.100)
				iter=iter+1
				xr0=xr
				const=1.18d7/fmdot**0.4d0/fr**1.4*fmtot**0.2d0*6.63d5**0.8d0 &
			        *fom**0.8d0
				fac=3.812d15/xr0/fr*fmtot+4.84d21*fom**2*xr0**2*fr**2
				dfacdx=-3.812d15/xr0**2/fr*fmtot+9.68d21*fom**2*xr0*fr**2
				f=xr0-const/fac**0.4d0
				dfdx=1.d0+0.40*const/fac**1.4d0*dfacdx
				dxr=-f/dfdx
				xr=xr0+dxr
			  end do
		!c         print 100,fom,xr,iter
		   100   format(1pe12.4,0pf8.3,i3)
			  fexp=(28.d0*wn+43.d0)/99.d0
			  raor=xr**fexp
		! c
			  fkn=fk*0.6666667d0*4.844d21*6.3031d11
			  djdt = fkn*fmdot*fom*fr**2*raor**wn
		! c      djdt = 2.539d58*fom**3*dsqrt(fr*6.96d10/(fmtot*1.989d33))
			  if (fk.lt.0.d0) then
				if (fom.le.1.13439d-5) djdt=6.226d46*fom**2.8558d0
				if (fom.gt.1.13439d-5) djdt=2.471d41*fom**1.736357d0
			  endif
			  if (djdt.eq.0.d0) djdt=1.d0
		if (mod(s% model_number, 10) == 0) write(*,*) 'Jdot is ', djdt
         
         ! djdt is dJ/dt, i.e. angular momentum is not specific
         
         do k = 1, s% nz
         	if ((s% q(k) < s% conv_mx1_top) .and. (s% q(k) > s% conv_mx1_bot)) then
	         	s% extra_jdot(k) = djdt / s% dm(k)
	         end if
         end do

      end subroutine kawaler_jdot
      
      subroutine extras_controls(s, ierr)
         type (star_info), pointer :: s
         integer, intent(out) :: ierr
         ierr = 0
         
         s% other_torque => kawaler_jdot
         
         ! this is the place to set any procedure pointers you want to change
         ! e.g., other_wind, other_mixing, other_energy  (see star_data.inc)
         
         
      end subroutine extras_controls
      
      
      integer function extras_startup(s, id, restart, ierr)
         type (star_info), pointer :: s
         integer, intent(in) :: id
         logical, intent(in) :: restart
         integer, intent(out) :: ierr
         ierr = 0
         extras_startup = 0
         if (.not. restart) then
            call alloc_extra_info(s)
         else ! it is a restart
            call unpack_extra_info(s)
         end if
      end function extras_startup
      

      ! returns either keep_going, retry, backup, or terminate.
      integer function extras_check_model(s, id, id_extra)
         type (star_info), pointer :: s
         integer, intent(in) :: id, id_extra
         extras_check_model = keep_going         
         if (.false. .and. s% star_mass_h1 < 0.35d0) then
            ! stop when star hydrogen mass drops to specified level
            extras_check_model = terminate
            write(*, *) 'have reached desired hydrogen mass'
            return
         end if
      end function extras_check_model


      integer function how_many_extra_history_columns(s, id, id_extra)
         type (star_info), pointer :: s
         integer, intent(in) :: id, id_extra
         how_many_extra_history_columns = 0
      end function how_many_extra_history_columns
      
      
      subroutine data_for_extra_history_columns(s, id, id_extra, n, names, vals, ierr)
         type (star_info), pointer :: s
         integer, intent(in) :: id, id_extra, n
         character (len=maxlen_history_column_name) :: names(n)
         real(dp) :: vals(n)
         integer, intent(out) :: ierr
     
                  
         !note: do NOT add these names to history_columns.list
         ! the history_columns.list is only for the built-in log column options.
         ! it must not include the new column names you are adding here.
         
         ierr = 0
      end subroutine data_for_extra_history_columns

      
      integer function how_many_extra_profile_columns(s, id, id_extra)
         type (star_info), pointer :: s
         integer, intent(in) :: id, id_extra
         how_many_extra_profile_columns = 0
      end function how_many_extra_profile_columns
      
      
      subroutine data_for_extra_profile_columns(s, id, id_extra, n, nz, names, vals, ierr)
         type (star_info), pointer :: s
         integer, intent(in) :: id, id_extra, n, nz
         character (len=maxlen_profile_column_name) :: names(n)
         real(dp) :: vals(nz,n)
         integer, intent(out) :: ierr
         integer :: k
         ierr = 0
         
         !note: do NOT add these names to profile_columns.list
         ! the profile_columns.list is only for the built-in profile column options.
         ! it must not include the new column names you are adding here.

         ! here is an example for adding a profile column
         !if (n /= 1) stop 'data_for_extra_profile_columns'
         !names(1) = 'beta'
         !do k = 1, nz
         !   vals(k,1) = s% Pgas(k)/s% P(k)
         !end do
         
      end subroutine data_for_extra_profile_columns
      

      ! returns either keep_going or terminate.
      ! note: cannot request retry or backup; extras_check_model can do that.
      integer function extras_finish_step(s, id, id_extra)
         type (star_info), pointer :: s
         integer, intent(in) :: id, id_extra
         integer :: ierr
         extras_finish_step = keep_going
         call store_extra_info(s)

         ! to save a profile, 
            ! s% need_to_save_profiles_now = .true.
         ! to update the star log,
            ! s% need_to_update_history_now = .true.
            
      end function extras_finish_step
      
      
      subroutine extras_after_evolve(s, id, id_extra, ierr)
         type (star_info), pointer :: s
         integer, intent(in) :: id, id_extra
         integer, intent(out) :: ierr
         ierr = 0
      end subroutine extras_after_evolve
      
      
      ! routines for saving and restoring extra data so can do restarts
         
         ! put these defs at the top and delete from the following routines
         !integer, parameter :: extra_info_alloc = 1
         !integer, parameter :: extra_info_get = 2
         !integer, parameter :: extra_info_put = 3
      
      
      subroutine alloc_extra_info(s)
         integer, parameter :: extra_info_alloc = 1
         type (star_info), pointer :: s
         call move_extra_info(s,extra_info_alloc)
      end subroutine alloc_extra_info
      
      
      subroutine unpack_extra_info(s)
         integer, parameter :: extra_info_get = 2
         type (star_info), pointer :: s
         call move_extra_info(s,extra_info_get)
      end subroutine unpack_extra_info
      
      
      subroutine store_extra_info(s)
         integer, parameter :: extra_info_put = 3
         type (star_info), pointer :: s
         call move_extra_info(s,extra_info_put)
      end subroutine store_extra_info
      
      
      subroutine move_extra_info(s,op)
         integer, parameter :: extra_info_alloc = 1
         integer, parameter :: extra_info_get = 2
         integer, parameter :: extra_info_put = 3
         type (star_info), pointer :: s
         integer, intent(in) :: op
         
         integer :: i, j, num_ints, num_dbls, ierr
         
         i = 0
         ! call move_int or move_flg    
         num_ints = i
         
         i = 0
         ! call move_dbl       
         
         num_dbls = i
         
         if (op /= extra_info_alloc) return
         if (num_ints == 0 .and. num_dbls == 0) return
         
         ierr = 0
         call star_alloc_extras(s% id, num_ints, num_dbls, ierr)
         if (ierr /= 0) then
            write(*,*) 'failed in star_alloc_extras'
            write(*,*) 'alloc_extras num_ints', num_ints
            write(*,*) 'alloc_extras num_dbls', num_dbls
            stop 1
         end if
         
         contains
         
         subroutine move_dbl(dbl)
            real(dp) :: dbl
            i = i+1
            select case (op)
            case (extra_info_get)
               dbl = s% extra_work(i)
            case (extra_info_put)
               s% extra_work(i) = dbl
            end select
         end subroutine move_dbl
         
         subroutine move_int(int)
            integer :: int
            i = i+1
            select case (op)
            case (extra_info_get)
               int = s% extra_iwork(i)
            case (extra_info_put)
               s% extra_iwork(i) = int
            end select
         end subroutine move_int
         
         subroutine move_flg(flg)
            logical :: flg
            i = i+1
            select case (op)
            case (extra_info_get)
               flg = (s% extra_iwork(i) /= 0)
            case (extra_info_put)
               if (flg) then
                  s% extra_iwork(i) = 1
               else
                  s% extra_iwork(i) = 0
               end if
            end select
         end subroutine move_flg
      
      end subroutine move_extra_info

      end module run_star_extras
      
